# -*- coding: utf-8 -*-
"""
Loop_list est une liste chainée circulaire. Elle permet dans la simulation de stocker les joueurs de façon pertinente.
Auteur du script: Nicolas VEYLON
"""

class Node:

    """Script similaire à celui du TD sur les liste chainées"""

    def __init__(self, instance):
        self._content = instance
        self._next = None
        
    """Setter et Getter"""
        
    @property
    def content(self):
        return self._content
    @content.setter
    def content(self, instance):
        self._content = instance
    @property
    def next(self):
        return self._next
    @next.setter
    def next(self, instance):
        self._next = instance
    
    def __str__(self):
        return "### Node ### \t {0}".format(self._content)


class Loop_list:

    """
    Liste chainée bouclée sur elle même
    """

    def __init__(self, first_content = None): #L'élément first est parfaitement arbitraire ici car la liste est circulaire
        self._first = Node(first_content)     #       ==> Il permet une insertion à un endroit arbitraire
        self._first._next = self._first       #La liste est toujours bouclée, même avec un seul élément.
                                              #La liste n'est jamais vide : au mieux elle contient un noeud contenant None
    def insert(self, instance):
        """
        Permet d'insérer un objet dans la liste.
        """
        new_node = Node(instance)
        if self._first._next == self._first:  #Cas où la liste ne contient qu'un seul élément
            new_node._next = self._first
            self._first._next = new_node
            if self._first._content == None:  #Sous-cas où la liste ne contenait qu'un noeud vide. Il est supprimé.
                self.remove(None)
        else:                                 #Cas où la liste contient plusieurs éléments
            new_node._next = self._first._next
            self._first._next = new_node
    
    def remove(self,instance):
        current = self._first        
        if self._first._content == instance: # on veut supprimer le premier élement (ie self._first)
            if self._first._next == self._first: #cas où la liste ne contient que cette élément
                self._first = Node(None)
                self._first._next = self._first
            else:                               #Cas où la liste contient plusieurs éléments                
                while current._next != self._first: #On se place au noeud juste avant l'élément à supprimer, ici self._first
                    current = current._next
                current._next = current._next._next
                self._first = self._first._next
        else:                               #on veut supprimer un élément différent du premier (ie self._first)
            while current._next._content != instance: #On se place au noeud juste avant l'élément à supprimer
                current = current._next
                if current == self._first:  #si l'élément à supprimer n'est pas dans la liste, on fait rien
                    return
            current._next = current._next._next
    
#    def change_first_content(self, instance):
#        self.first.content = instance
    
    @property
    def first(self):
        return self._first 
    
#    def Getter_first_content(self):
#        return self.first.content
     
    def __str__(self):            
        tab = [self._first._content]
        current = self._first._next
        while current !=self._first:
            tab.append(current._content)
            current = current._next
        return str(tab)
        
if __name__ == "__main__":
    chain = Loop_list()
    assert chain.__str__() == "[None]"   #vérification de la création
    chain.insert("pop")
    assert chain.__str__() == "['pop']"   #vérification de l'écrasement de None
    chain.remove("pop")
    assert chain.__str__() == "[None]"    #vérification de la liste vide suite à la suppression de l'unique élément    
    chain.insert("pop")    
    chain.insert("pop pop pop")
    assert chain.first._next._next == chain._first    #vérification du bouclage
    chain.insert("pop pop")
    chain.remove('pop')
    assert chain.__str__() == "['pop pop', 'pop pop pop']"  #vérification de la suppression
    assert chain._first._content == 'pop pop'   #vérification de la modification de first lors de la suppresion
    print("Tout les systèmes sont verts")