# -*- coding: utf-8 -*-

import numpy as np
import random
from Objet import *
from Fct_event import *
from save import save

"""
Classe Personne, ce sont les joueurs (vivant et zombie) de la simulation.
Les instances sont stockées dans la variable d'instance population de Case, dans la liste (Loop_list) des joueurs players et dans spawn
"""

spawn = [] #Liste des nouveaux joueurs apparut lors du tour, qui se vide à chaque début de tour

################################Classe Personne###########################################
class Personne(object):
    "INIT"
    def __init__ (self,pos): #pos est un objet Case
        self._pos=pos
        self._isdead = False #si isdead == True, on procède à l'élimination du joueur
        self._respawn = None #sert de compte à rebour pour les zombies. Il décrémente à chaque tour. Tant qu'il n'est pas à 0, le zombie ne joue pas.
        self._equi=[]        #liste d'instances Objet possédés par la personne.
        spawn.append(self)   #ajoute l'instance créée à spawn
    
    "GETTER"
    @property
    def pos(self):
        return self._pos
    @property
    def isdead(self):
        return self._isdead
    @property
    def respawn(self):
        return self._respawn
    @property
    def equi(self):
        return self._equi
        
    "SETTER"
    @pos.setter    
    def pos(self,position):
        self._pos=position
    @isdead.setter
    def isdead(self,True_or_False):
        self._isdead = True_or_False
    @equi.setter
    def equi(self, modif):
        self._equi = modif
    @respawn.setter
    def respawn(self, modif):
        self._respawn = modif
    
    def Modif_equi(self,modif): #modif = array([[instance objet,0],[instance objet, 1]...])
        """        
        Permet d'ajouter ou supprimer un ensemble d'objets Objet de la variable d'instance equi
        Auteur: Sophie TUTON
        """
        for elt in modif:   #0 signifie "supprimer", 1 signifie "ajouter"
            if int(elt[1])==0:
                try:
                   self._equi.remove(elt[0])
                   self._pos.loot.append(elt[0])
                   elt[0].pos = self._pos
                   
                except: #On traite le cas où la personne n'a pas d'équipemment
                   pass
            else:
                self._equi.append(elt[0])
                self._pos.loot.remove(elt[0])
                
    
   

################################Sous-classe: Zombie et Vivant###########################################
class Zombie(Personne):
    nbZ = 0  #variable de classe servant à tracer l'évolution du nombre de Zombie dans le temps
    "INIT"    
    def __init__(self,pos):
        super().__init__(pos)
        self._respawn = 3        #Délai de respawn. Ne joue pas tant qu'il n'est pas descendu à 0.
        Zombie.nbZ += 1                 #          ==> Les vivants qui meurent mettent un certain temps à se relever.
    
    
    def tour(self,carte):
        """
        Méthode principale de la classe Zombie et génère le comportement du zombie pendant un tour
        Un zombie ne cherche qu'à manger les vivants
        Auteur: Nicolas VEYLON
        """
        choix = self.observation_decision(carte)
        self.deplacement(choix, carte)
        
    
    "OBSERVATION"
    def observation_decision(self, carte):
        """
        Retourne la case choisie par le Zombie.
        Un zombie ne cherche qu'à manger les vivants dans la case la plus proche
        Si il n'y a pas de viavnt dans sa zone d'observation, alors il a une certaine probabilité d'errer, i.e. de se déplacer d'une case autour.
        Auteur: Nicolas VEYLON et Sophie TUTON 
        """
        choix = None
        #On choisit la case la plus proche où il y a des vivants dans la zone d'observation.
        distance=5
        for i in range (-2,3):
            for j in range(-2,3):
                try:
                    for individu in carte[self._pos.pos[0]+i,self.pos.pos[1]+j].population:
                        if type(individu) is not LambdaZ and type(individu) is not Tank  and abs(i)+abs(j) < distance:
                            distance=abs(i)+abs(j)
                            choix= individu.pos
                except:
                    pass
        # Si le choix n'a pas été définit avant, i.e. qu'il n'y a pas de vivant dans la zone d'observation, alors le zombie erre       
        if choix == None:
            choix=self._pos
            if random.random() < 0.5: 
                r0=round(random.random()*2-1)
                r1=round(random.random()*2-1)
                try:
                    choix = carte[self.pos.pos[0]+r0,self.pos.pos[0]+r1]
                except:
                    pass
            elif random.random() < 0.5:
                choix = carte[len(carte)//2, len(carte)//2]
        return choix
    
    "DEPLACEMENT"
    def deplacement(self,choix,carte):
        """
        Cette méthode permet le déplacement des Zombies en fonction du choix retourné par observation
        Il va se déplacer sur la case choisi en combattant avec chacun des Vivants qu'il croise.
        Auteur: Nicolas VEYLON et Sophie TUTON
        """
        #Phase 1:  combat à la case initiale
        for individu in self._pos.population:
            if type(individu) is not LambdaZ and type(individu) is not Tank:
                while not individu.isdead:
                    if individu.combat(self) == 'esquive': #on combat
                        break
                    if self._isdead:
                        return
        if self._pos.pos == choix._pos:
            print((self.__class__.__name__ + " reste en " + str(self.pos.pos)))
            save((self.__class__.__name__ + " reste en " + str(self.pos.pos)))
            return
            
        #Phase 2: le déplacement. Le déplacement s'effectue selon la coordonnée la plus éloignée du but :
        delta_x = choix.pos[0] - self._pos.pos[0]
        delta_y = choix.pos[1] - self._pos.pos[1]
        if abs(delta_x) >= abs(delta_y):
            if delta_x >= 0:
                new_coor = (self._pos.pos[0]+1,self._pos.pos[1])
            else:
                new_coor = (self._pos.pos[0]-1,self._pos.pos[1])
                    
        else:
            if delta_y >= 0:
                new_coor = (self._pos.pos[0],self._pos.pos[1]+1)
            else:
                new_coor = (self._pos.pos[0],self._pos.pos[1]-1)
        self._pos.population.remove(self) # On supprime le zombie de la variable d'instance population de la case initiale
        self._pos = carte[new_coor] # On définit sa nouvelle variable d'instance pos
        carte[new_coor].population.append(self) #On rajoute le zombie dans la variable d'instance population de la nouvelle case 
        print("{0} s'est déplacé en {1}".format(self.__class__.__name__, self._pos.pos))
        save("{0} s'est déplacé en {1}".format(self.__class__.__name__, self._pos.pos))
        
        #Phase 3: Combat sur la case finale
        for individu in self._pos.population:
            if type(individu) is not LambdaZ and type(individu) is not Tank:
                while not individu.isdead:
                    if individu.combat(self) == 'esquive': #on combat
                        break
                    if self._isdead:
                        return
            
class Vivant(Personne):
    "INIT"
    nbV = 0 #variable de classe servant à tracer l'évolution du nombre de Vivants dans le temps
    def __init__(self,pos):
        super(). __init__(pos)
        self._respawn = 0
        self._faim=random.randint(60,99)
        self._moral=random.randint(7,10)
        self._courage=None
        self._force=None
        self._surv=None
        self._maitrise=None
        self._sick = False
        Vivant.nbV +=1
    
    "GETTER"   
    @property
    def faim(self):
        return(self._faim)
    @property
    def moral(self):
        return(self._moral)
    @property
    def courage(self):
        return(self._cour)
    @property
    def force(self):
        return self._force
    @property
    def surv(self):
        return self._force
    @property
    def maitrise(self):
        return self._maitrise
    @property
    def respawn(self):
        return(self._respawn)
    @property
    def sick(self):
        return(self._sick)
        
    "SETTER"
    @faim.setter
    def faim(self,modif):
        self._faim = modif
    @moral.setter
    def moral(self,modif):
        if 10>=modif >=0:
            self._moral = modif
    @courage.setter
    def courage(self,modif):
        self._courage = modif
    @force.setter
    def force(self,modif):
        self._force = modif
    @respawn.setter
    def respawn(self,modif):
        self._respawn = modif
    @sick.setter
    def sick(self, modif):
        self._sick = modif
        
    
    def combat(self,zombie):
        """
        Permet de simuler les combats entre Vivant et Zombie. 
        Auteur: Nicolas VEYLON et Sophie TUTON
        """
        firearm=False
        for elt in self._equi:
            if type(elt) is Arme_a_feu:
                firearm=True
        if firearm: #Si le vivant possède une arme à feu, il tente d'abattre le zombie à distance
            jet_tir=random.random()
            if jet_tir <0.08*self._maitrise:
                print("{0} s'est pris une balle".format(zombie.__class__.__name__))
                save("{0} s'est pris une balle".format(zombie.__class__.__name__))
                zombie.degat()
                return
        jet_esquive=random.random() #Le vivant tente une esquive
        if jet_esquive<0.06*self._surv+0.03*self._force:
            print("{0} a esquive".format(self.__class__.__name__))
            save("{0} a esquive".format(self.__class__.__name__))
            return 'esquive'
        meleearm=0
        for elt in self._equi:
            if type(elt) is Arme_de_melee:
                meleearm=1
        jet_melee=random.random() #Le vivant n'a plus le choix. Il combat au corps à corps.
                                  #Il se battra mieux avec une arme de melee
        if jet_melee*(1-0.1*max(0,case_v(self.pos)-case_z(self.pos)))  <  (0.08*self._force+0.06*self._surv)/(2-meleearm) and random.random() < 0.85:
            zombie.degat()
        else:
            self.degat()
     
    def degat(self):
         """
         Permet de zombifier les Vivants lorsqu'ils perdent un combat ou que leur faim tombe à 0
         La zombification d'un Vivant engendre une diminution de moral chez les vivants ce trouvant sur la même case
         Auteur: Nicolas VEYLON et Sophie TUTON
         """
         print("{0} meurt".format(self.__class__.__name__))
         save("{0} meurt".format(self.__class__.__name__))
         self._isdead = True #La variable d'instance isdead devient True
         
         #diminution du moral des Vivants         
         for pers in self.pos.population:
             if type(pers) is not LambdaZ and type(pers) is not Tank and pers!=self:
                 pers.moral -=1
                 print("Baisse de moral chez "+ str(pers.__class__.__name__))
                 save("Baisse de moral chez "+ str(pers.__class__.__name__))
                 
         #Traitement de la zombification
         self._pos.Modif_pop(np.array([[self,0]]))
         zombie = LambdaZ(self._pos)
         zombie.equi = self.equi
         self.equi = []
         self._pos.Modif_pop(np.array([[zombie, 1]]))
         Vivant.nbV -= 1

             
    def tour(self, carte):
        """
        Il s'agit de la méthode principale des Vivants.
        Elle traite la faim, la maladie, la prise de décision et le déplacement.
        Auteur: Nicolas VEYLON et Sophie TUTON
        """
        #Phase 1: la Faim
        self.faim -= 5
        if self.faim < 50:
            for elt in self._equi:
                if type(elt) is Nourriture:
                    self.equi.remove(elt)
                    self.faim += 30
                    print("{0} mange".format(self.__class__.__name__))
                    save("{0} mange".format(self.__class__.__name__))
                    break
        if self._faim < 0:
            print("{0} meurt de faim".format(self.__class__.__name__))  
            save("{0} meurt de faim".format(self.__class__.__name__))
            self.degat()
            return
            
        #Phase 2: la maladie
        if random.random() < 0.040:
            maladie(self)
        if self.sick:
            for elt in self._equi:
                if type(elt) is Medicament:
                    soin(self)
                    self._equi.remove(elt)
                    break
                
        #☻Phase 3: L'observation, la décision et le déplacement    
        choix = self.observation_decision(carte)
        self.deplacement(choix,carte)
        if choix != None and self._pos.pos == choix.pos.pos and not self._isdead:
            self.Modif_equi(np.array([[choix,1]]))
            print("{0} a récupéré {1}".format(self.__class__.__name__, choix.__class__.__name__))
            save("{0} a récupéré {1}".format(self.__class__.__name__, choix.__class__.__name__))
        
    '''OBSERVATION ET DECISION'''
    def observation_decision(self,carte):
        """
        Chaque objet est associé à un interet en fonction de la situation. 
        Le Vivant choisira l'objet le plus interressant pour lui sur la case d'interet maximal.
        Elle renvoie l'instance objet correspante.
        Auteur: Nicolas VEYLON et Sophie TUTON
        """
        MZ = np.zeros((5,5)) #Cette matrice représente le nombre de zombie autour du vivant. Elle est donc centrée sur lui
        for i in range(-2,3):
            for j in range(-2,3):
                if 0 <= self.pos.pos[0]+i < carte.shape[0] and 0 <= self.pos.pos[1]+j < carte.shape[1]:
                    n=0 
                    for personne in carte[self.pos.pos[0]+i,self.pos.pos[1]+j].population:
                        if type(personne) is LambdaZ:
                            n+=1
                        elif type(personne) is Tank:
                            n+=3
                    MZ[i+2,j+2]=n
                else:
                    MZ[i+2,j+2] = None
        
        Cg, Ch, Cd, Cb = 0,0,0,0 #Comptage des zombies dans chaque "cone" de la matrice (voir rapport)
        for i in range(5):
            for j in range(5):
                
                if type(MZ[i,j]) is float:
                    if i+j <= 5 and i>=j:
                        Cg += MZ[i,j]
                    if i+j <= 5 and i<=j:
                        Ch += MZ[i,j]
                    if i+j >= 5 and i<=j:
                        Cd += MZ[i,j]
                    if i+j >= 5 and i>=j:
                        Cb += MZ[i,j]
        if self.courage < 3: #Definition du nombre de zombie max que le joueur se ent capable d'affronter
            maxZ=0
        elif self.courage <=5:
            maxZ=1
        elif self.courage <= 6:
            maxZ=2
        elif self.courage <= 7:
            maxZ=3
        elif self.courage<= 9:
            maxZ=5
        else:
            maxZ=np.inf
        choix=None #Objet choisi par le vivant
        Ichoix=0 #Interet de la case la plus interressante
        for i in range(5): #Choix de l'objet. POur les details voir interet()
            for j in range(5):
                if i+j <= 5 and i>=j and Cg<=maxZ: #On ne traite que les cases des cones qui n'ont pas été éliminés
                    try:
                        (I,loot)= self.interet(carte[self.pos.pos[0]+i-2,self.pos.pos[1]+j-2])
                        if I> Ichoix:
                            Ichoix,choix=I,loot
                    except:
                        pass
                if i+j <= 5 and i<=j and Ch<=maxZ:
                    try:
                        (I,loot)= self.interet(carte[self.pos.pos[0]+i-2,self.pos.pos[1]+j-2])
                        if I> Ichoix:
                            Ichoix,choix=I,loot
                    except:
                        pass

                if i+j >= 5 and i<=j and Cd<=maxZ:
                    try:                        
                        (I,loot)= self.interet(carte[self.pos.pos[0]+i-2,self.pos.pos[1]+j-2])
                        if I> Ichoix:
                            Ichoix,choix=I,loot
                    except:
                        pass

                if i+j >= 5 and i>=j and Cb<=maxZ:
                    try:
                        (I,loot)= self.interet(carte[self.pos.pos[0]+i-2,self.pos.pos[1]+j-2])
                        if I> Ichoix:
                            Ichoix,choix=I,loot
                    except:
                        pass
        return(choix)
     
    def interet(self,case):
        """
        Cette méthode calcule l'interet de la case mise en paramètre
        Elle retourne également l'objet le plus interressant sur cette case 
        Auteur: Nicolas VEYLON et Sophie TUTON        
        """
        I=0
        presence_Nourriture,presence_Medicament, presence_Arme_de_melee, presence_Arme_a_feu=False,False,False,False
        #==> La redondance des objets sur la case n'est pas interressant
        Possession={'Nourriture':False,'Arme_a_feu':False,'Arme_de_melee':False,'Medicament':False}
        #==> La possession ou non du même objet le rend moins interressant        
        nb_Nourriture = 0        
        for objet in self.equi:
            Possession[str(type(objet))] = True
            if type(objet) is Nourriture:
                nb_Nourriture += 1
        choix = None
        Imax = 0 #interet de l'objet
        for elt in case.loot:
            I_old = I
            if type(elt) is Nourriture and not presence_Nourriture and nb_Nourriture <= 2: #Les personnes ne peuvent pas avoir un nombre illimité de nourriture sur eux
                if self._faim < 20:
                    I+= 300*(1-0.5*Possession["Nourriture"])
                elif self._faim < 50:
                    I+= 50*(1-0.5*Possession["Nourriture"])
                else:
                    I+=100-self._faim*(1-0.5*Possession["Nourriture"])
                presence_Nourriture= True
            elif  type(elt) is Medicament and self._sick and not presence_Medicament:
                I += 120
                presence_Medicament=True
            elif  type(elt) is Arme_a_feu and not presence_Arme_a_feu and not Possession["Arme_a_feu"]:
                I += 70
                presence_Arme_a_feu = True
            elif  type(elt) is Arme_de_melee and not presence_Arme_de_melee and not Possession["Arme_de_melee"]:
                I += 90
                presence_Arme_de_melee = True
            if Imax < I-I_old: #Permet de chosiir l'objet le plus interressant : I-I_old = I_objet
                Imax = I-I_old
                choix = elt
        for pers in case.population: #Les perosnnes ave beaucoup de moral se rassemblent, tandis que ceux qui n'en ont pas beaucoup s'évitent
            if type(pers) is not LambdaZ and type(pers) is not Tank and pers!=self:
                I+=(self._moral-5)*4
                break
            

        return(I,choix)
        
    def deplacement(self,choix, carte):
        """
        Cette méthode permet au vivant de se déplacer vers l'objet choisit précédemment avec observation_decision()
        Sur chaque case parcourue, si il y a un zombie, il le combat.
        Auteur: Nicolas VEYLON et SOphie TUTON
        """
        if type(choix) is not Nourriture and self.faim <= 30: #vers le supermarché s'ils ont faim
                destination = (len(carte)//5,len(carte[0])//5)
        else:
            if choix == None: #Si aucun objet a été choisi, on les fait aller vers le centre de la carte
                    destination = (carte.shape[0]//2, carte.shape[1]//2)
            else:
                destination = choix.pos.pos
        for individu in self._pos.population: #combat s'il y a un ou plusieurs zombies sur la case de départ.
                if type(individu) is LambdaZ or type(individu) is Tank:
                    while not individu.isdead:
                        if self.combat(individu) == 'esquive':
                            break
                        if self._isdead:
                            return
            
        for i in range(self.moral//4+1) :           
            if self._pos.pos == destination:
                print((self.__class__.__name__ + " reste en " + str(self.pos.pos)))
                save((self.__class__.__name__ + " reste en " + str(self.pos.pos)))
                break
            delta_x = destination[0] - self._pos.pos[0]
            delta_y = destination[1] - self._pos.pos[1]
            if abs(delta_x) >= abs(delta_y):
                if delta_x >= 0: #deplacement selon la coordonnée la plus éloignée du but
                    new_coor = (self._pos.pos[0]+1,self._pos.pos[1])
                else:
                    new_coor = (self._pos.pos[0]-1,self._pos.pos[1])
                    
            else:
                if delta_y >= 0:
                    new_coor = (self._pos.pos[0],self._pos.pos[1]+1)
                else:
                    new_coor = (self._pos.pos[0],self._pos.pos[1]-1)
            self._pos.population.remove(self)
            self._pos = carte[new_coor]
            carte[new_coor].population.append(self)
            print("{0} s'est déplacé en {1}".format(self.__class__.__name__, self._pos.pos))
            save("{0} s'est déplacé en {1}".format(self.__class__.__name__, self._pos.pos))
            for individu in self._pos.population: #combat s'il y a un ou des zombies sur la case
                if type(individu) is LambdaZ or type(individu) is Tank:
                    while not individu.isdead:
                        if self.combat(individu) == 'esquive':
                            break
                        if self._isdead:
                            return
            

        
    def __str__(self):
        return "{0}: position {1},equipement {2},faim {3},moral {4}, Courage {5}, Maitrise {6}".format(self.__class__.__name__,self._pos.pos,self._equi,self._faim,self._moral,self._cour,self._maitrise)
            #A modifier avec les nouveaux attributs
################################Sous sous--classe###########################################


class Tank(Zombie):
    
    "INIT"
    def __init__(self,pos):
        super().__init__(pos)
        self._etat=4
        
    "GETTER/SETTER"
    @property    
    def etat(self):
        return self._etat

    def degat(self):
        """
        Permet de tuer le Tank lorsqu'il perd un combat et qu'il lui reste plus qu'un seul point d'état. Sinon elle décrémente état de 1.
        La mort d'un Zombie Tank engendre une augmentation de moral chez les vivants ce trouvant sur la même case
        Auteur: Nicolas VEYLON et Sophie TUTON
        """
        self._etat = self._etat-1
        if self._etat <= 0:
            #déclaration du ZOmbie comme mort.
            print("{0} meurt".format(self.__class__.__name__))
            save("{0} meurt".format(self.__class__.__name__))
            self._isdead = True
            #augmentation du moral chez les vivants :
            for pers in self.pos.population:
             if type(pers) is not LambdaZ and type(pers) is not Tank:
                 pers.moral +=1
                 print("Augmentation de moral chez "+ str(pers.__class__.__name__))
                 save("Augmentation de moral chez "+ str(pers.__class__.__name__))
            self._pos.Modif_pop(np.array([[self,0]]))
            T=[]
            for obj in self.equi: #depot de l'équipement sur la case.
                T.append([obj,1])
            self.pos.Modif_loot(np.array(T))
            self.equi = []
            Zombie.nbZ -= 1
            
        
    "STR"
    def __str__(self):
        return "{0}; position {1} ; equipement:{2}; etat : {3}".format(self.__class__.__name__, self._pos.pos,self._equi,self._etat)      
            #A modifier avec les nouveaux attributs
    


class LambdaZ(Zombie):

    "INIT"    
    def __init__(self,pos):
        super().__init__(pos)
        
    
    def degat(self):
        """
        Permet de tuer le Zombie lorsqu'il perd un combat;
        La mort d'un Zombie engendre une augmentation de moral chez les vivants ce trouvant sur la même case
        Auteur: Nicolas VEYLON et Sophie TUTON
        """        
        #Déclaration du zombie comme mort
        print("{0} meurt".format(self.__class__.__name__))
        save("{0} meurt".format(self.__class__.__name__))
        self._isdead = True
        for pers in self.pos.population:
             if type(pers) is not LambdaZ and type(pers) is not Tank:
                 pers.moral +=1
                 print("Augmentation de moral chez "+ str(pers.__class__.__name__))
                 save("Augmentation de moral chez "+ str(pers.__class__.__name__))
        self._pos.Modif_pop(np.array([[self,0]]))
        T=[]
        for obj in self.equi: #depot de l'équipement du zombie sur la case
            T.append([obj,1])
        self.pos.Modif_loot(np.array(T))
        self.equi = []   
        Zombie.nbZ -= 1
    "STR"
    def __str__(self):
        return "{0}; position {1} ; equipement:{2}".format(self.__class__.__name__, self._pos.pos,self._equi)  
        #A modifier avec les nouveaux attributs


class Enfant(Vivant):
    "INIT"
    def __init__(self,pos):
        super().__init__(pos)
        self._cour=random.randint(1,6)
        self._force=random.randint(1,5)
        self._surv=random.randint(0,1)
        self._maitrise=random.randint(0,3)
 
       
class Medecin(Vivant):
    "INIT"
    def __init__(self,pos):
        super().__init__(pos,)
        self._cour=random.randint(5,8)
        self._force=random.randint(3,7)
        self._surv=random.randint(0,8)
        self._maitrise=random.randint(0,6)
        
class Religieux_ext(Vivant):

    "INIT"
    def __init__(self,pos):
        super().__init__(pos)
        self._cour=10
        self._force=random.randint(2,6)
        self._surv=random.randint(0,2)
        self._maitrise=0
        
class Militaire(Vivant):
    "INIT"
    def __init__(self,pos):
        super().__init__(pos)
        self._cour=random.randint(8,10)
        self._force=random.randint(6,10)
        self._surv=random.randint(6,10)
        self._maitrise=random.randint(8,10)
        
class LambdaV(Vivant):
    "INIT"
    def __init__(self,pos):
        super().__init__(pos)
        self._cour=random.randint(0,9)
        self._force=random.randint(3,7)
        self._surv=random.randint(0,8)
        self._maitrise=random.randint(0,6)
        
        

"""fonctions utiles"""

def case_v(case):
        n = 0
        for individu in case.population:
            if type(individu) is not LambdaZ and type(individu) is not Tank:
                n += 1
        return n
        
def case_z(case):
    n = 0
    for individu in case.population:
        if type(individu) is LambdaZ or type(individu) is Tank:
            n += 1
    return n
        